@@ -16,23 +16,19 @@
     2. Type conversion for read and write (currently untested)
 """
 
-from __future__ import absolute_import
-
+import pathlib
 import sys
-
-import six
-
 import numpy as np
-
 import platform
-
 import pytest
 
 from .common import ut, TestCase
+from .data_files import get_data_file_path
 from h5py import File, Group, Dataset
 from h5py._hl.base import is_empty_dataspace
 from h5py import h5f, h5t
 import h5py
+import h5py._hl.selections as sel
 
 
 class BaseDataset(TestCase):
@@ -52,9 +48,9 @@ class TestRepr(BaseDataset):
     def test_repr_open(self):
         """ repr() works on live and dead datasets """
         ds = self.f.create_dataset('foo', (4,))
-        self.assertIsInstance(repr(ds), six.string_types)
+        self.assertIsInstance(repr(ds), str)
         self.f.close()
-        self.assertIsInstance(repr(ds), six.string_types)
+        self.assertIsInstance(repr(ds), str)
 
 
 class TestCreateShape(BaseDataset):
@@ -73,6 +69,11 @@ class TestCreateShape(BaseDataset):
         dset = self.f.create_dataset('foo', (1,))
         self.assertEqual(dset.shape, (1,))
 
+    def test_create_integer(self):
+        """ Create a size-1 dataset with integer shape"""
+        dset = self.f.create_dataset('foo', 1)
+        self.assertEqual(dset.shape, (1,))
+
     def test_create_extended(self):
         """ Create an extended dataset """
         dset = self.f.create_dataset('foo', (63,))
@@ -82,6 +83,15 @@ class TestCreateShape(BaseDataset):
         self.assertEqual(dset.shape, (6, 10))
         self.assertEqual(dset.size, (60))
 
+    def test_create_integer_extended(self):
+        """ Create an extended dataset """
+        dset = self.f.create_dataset('foo', 63)
+        self.assertEqual(dset.shape, (63,))
+        self.assertEqual(dset.size, 63)
+        dset = self.f.create_dataset('bar', (6, 10))
+        self.assertEqual(dset.shape, (6, 10))
+        self.assertEqual(dset.size, (60))
+
     def test_default_dtype(self):
         """ Confirm that the default dtype is float """
         dset = self.f.create_dataset('foo', (63,))
@@ -106,6 +116,12 @@ class TestCreateShape(BaseDataset):
                                      dtype=np.dtype('complex256'))
         self.assertEqual(dset.dtype, np.dtype('complex256'))
 
+    def test_name_bytes(self):
+        dset = self.f.create_dataset(b'foo', (1,))
+        self.assertEqual(dset.shape, (1,))
+
+        dset2 = self.f.create_dataset(b'bar/baz', (2,))
+        self.assertEqual(dset2.shape, (2,))
 
 class TestCreateData(BaseDataset):
 
@@ -143,11 +159,44 @@ class TestCreateData(BaseDataset):
         with self.assertRaises(ValueError):
             Dataset(self.f['/'].id)
 
-    @ut.expectedFailure
+    def check_h5_string(self, dset, cset, length):
+        tid = dset.id.get_type()
+        assert isinstance(tid, h5t.TypeStringID)
+        assert tid.get_cset() == cset
+        if length is None:
+            assert tid.is_variable_str()
+        else:
+            assert not tid.is_variable_str()
+            assert tid.get_size() == length
+
     def test_create_bytestring(self):
         """ Creating dataset with byte string yields vlen ASCII dataset """
-        # there was no test here!
-        self.assertEqual(True, False)
+        def check_vlen_ascii(dset):
+            self.check_h5_string(dset, h5t.CSET_ASCII, length=None)
+        check_vlen_ascii(self.f.create_dataset('a', data=b'abc'))
+        check_vlen_ascii(self.f.create_dataset('b', data=[b'abc', b'def']))
+        check_vlen_ascii(self.f.create_dataset('c', data=[[b'abc'], [b'def']]))
+        check_vlen_ascii(self.f.create_dataset(
+            'd', data=np.array([b'abc', b'def'], dtype=object)
+        ))
+
+    def test_create_np_s(self):
+        dset = self.f.create_dataset('a', data=np.array([b'abc', b'def'], dtype='S3'))
+        self.check_h5_string(dset, h5t.CSET_ASCII, length=3)
+
+    def test_create_strings(self):
+        def check_vlen_utf8(dset):
+            self.check_h5_string(dset, h5t.CSET_UTF8, length=None)
+        check_vlen_utf8(self.f.create_dataset('a', data='abc'))
+        check_vlen_utf8(self.f.create_dataset('b', data=['abc', 'def']))
+        check_vlen_utf8(self.f.create_dataset('c', data=[['abc'], ['def']]))
+        check_vlen_utf8(self.f.create_dataset(
+            'd', data=np.array(['abc', 'def'], dtype=object)
+        ))
+
+    def test_create_np_u(self):
+        with self.assertRaises(TypeError):
+            self.f.create_dataset('a', data=np.array([b'abc', b'def'], dtype='U3'))
 
     def test_empty_create_via_None_shape(self):
         self.f.create_dataset('foo', dtype='f')
@@ -157,6 +206,99 @@ class TestCreateData(BaseDataset):
         self.f.create_dataset('foo', data=h5py.Empty(dtype='f'))
         self.assertTrue(is_empty_dataspace(self.f['foo'].id))
 
+    def test_create_incompatible_data(self):
+        # Shape tuple is incompatible with data
+        with self.assertRaises(ValueError):
+            self.f.create_dataset('bar', shape=4, data= np.arange(3))
+
+
+class TestReadDirectly:
+
+    """
+        Feature: Read data directly from Dataset into a Numpy array
+    """
+
+    @pytest.mark.parametrize(
+        'source_shape,dest_shape,source_sel,dest_sel',
+        [
+            ((100,), (100,), np.s_[0:10], np.s_[50:60]),
+            ((70,), (100,), np.s_[50:60], np.s_[90:]),
+            ((30, 10), (20, 20), np.s_[:20, :], np.s_[:, :10]),
+            ((5, 7, 9), (6,), np.s_[2, :6, 3], np.s_[:]),
+        ])
+    def test_read_direct(self, writable_file, source_shape, dest_shape, source_sel, dest_sel):
+        source_values = np.arange(np.product(source_shape), dtype="int64").reshape(source_shape)
+        dset = writable_file.create_dataset("dset", source_shape, data=source_values)
+        arr = np.full(dest_shape, -1, dtype="int64")
+        expected = arr.copy()
+        expected[dest_sel] = source_values[source_sel]
+
+        dset.read_direct(arr, source_sel, dest_sel)
+        np.testing.assert_array_equal(arr, expected)
+
+    def test_no_sel(self, writable_file):
+        dset = writable_file.create_dataset("dset", (10,), data=np.arange(10, dtype="int64"))
+        arr = np.ones((10,), dtype="int64")
+        dset.read_direct(arr)
+        np.testing.assert_array_equal(arr, np.arange(10, dtype="int64"))
+
+    def test_empty(self, writable_file):
+        empty_dset = writable_file.create_dataset("edset", dtype='int64')
+        arr = np.ones((100,), 'int64')
+        with pytest.raises(TypeError):
+            empty_dset.read_direct(arr, np.s_[0:10], np.s_[50:60])
+
+    def test_wrong_shape(self, writable_file):
+        dset = writable_file.create_dataset("dset", (100,), dtype='int64')
+        arr = np.ones((200,))
+        with pytest.raises(TypeError):
+            dset.read_direct(arr)
+
+    def test_not_c_contiguous(self, writable_file):
+        dset = writable_file.create_dataset("dset", (10, 10), dtype='int64')
+        arr = np.ones((10, 10), order='F')
+        with pytest.raises(TypeError):
+            dset.read_direct(arr)
+
+class TestWriteDirectly:
+
+    """
+        Feature: Write Numpy array directly into Dataset
+    """
+
+    @pytest.mark.parametrize(
+        'source_shape,dest_shape,source_sel,dest_sel',
+        [
+            ((100,), (100,), np.s_[0:10], np.s_[50:60]),
+            ((70,), (100,), np.s_[50:60], np.s_[90:]),
+            ((30, 10), (20, 20), np.s_[:20, :], np.s_[:, :10]),
+            ((5, 7, 9), (6,), np.s_[2, :6, 3], np.s_[:]),
+        ])
+    def test_write_direct(self, writable_file, source_shape, dest_shape, source_sel, dest_sel):
+        dset = writable_file.create_dataset('dset', dest_shape, dtype='int32', fillvalue=-1)
+        arr = np.arange(np.product(source_shape)).reshape(source_shape)
+        expected = np.full(dest_shape, -1, dtype='int32')
+        expected[dest_sel] = arr[source_sel]
+        dset.write_direct(arr, source_sel, dest_sel)
+        np.testing.assert_array_equal(dset[:], expected)
+
+    def test_empty(self, writable_file):
+        empty_dset = writable_file.create_dataset("edset", dtype='int64')
+        with pytest.raises(TypeError):
+            empty_dset.write_direct(np.ones((100,)), np.s_[0:10], np.s_[50:60])
+
+    def test_wrong_shape(self, writable_file):
+        dset = writable_file.create_dataset("dset", (100,), dtype='int64')
+        arr = np.ones((200,))
+        with pytest.raises(TypeError):
+            dset.write_direct(arr)
+
+    def test_not_c_contiguous(self, writable_file):
+        dset = writable_file.create_dataset("dset", (10, 10), dtype='int64')
+        arr = np.ones((10, 10), order='F')
+        with pytest.raises(TypeError):
+            dset.write_direct(arr)
+
 
 class TestCreateRequire(BaseDataset):
 
@@ -176,6 +318,20 @@ class TestCreateRequire(BaseDataset):
         dset2 = self.f.require_dataset('foo', (10, 3), 'f')
         self.assertEqual(dset, dset2)
 
+    def test_create_1D(self):
+        """ require_dataset with integer shape yields existing dataset"""
+        dset = self.f.require_dataset('foo', 10, 'f')
+        dset2 = self.f.require_dataset('foo', 10, 'f')
+        self.assertEqual(dset, dset2)
+
+        dset = self.f.require_dataset('bar', (10,), 'f')
+        dset2 = self.f.require_dataset('bar', 10, 'f')
+        self.assertEqual(dset, dset2)
+
+        dset = self.f.require_dataset('baz', 10, 'f')
+        dset2 = self.f.require_dataset(b'baz', (10,), 'f')
+        self.assertEqual(dset, dset2)
+
     def test_shape_conflict(self):
         """ require_dataset with shape conflict yields TypeError """
         self.f.create_dataset('foo', (10, 3), 'f')
@@ -195,6 +351,13 @@ class TestCreateRequire(BaseDataset):
         with self.assertRaises(TypeError):
             self.f.require_dataset('foo', (10, 3), 'S10')
 
+    def test_dtype_exact(self):
+        """ require_dataset with exactly dtype match """
+
+        dset = self.f.create_dataset('foo', (10, 3), 'f')
+        dset2 = self.f.require_dataset('foo', (10, 3), 'f', exact=True)
+        self.assertEqual(dset, dset2)
+
     def test_dtype_close(self):
         """ require_dataset with convertible type succeeds (non-strict mode)
         """
@@ -215,11 +378,21 @@ class TestCreateChunked(BaseDataset):
         dset = self.f.create_dataset('foo', shape=(100,), chunks=(10,))
         self.assertEqual(dset.chunks, (10,))
 
+    def test_create_chunks_integer(self):
+        """ Create via chunks integer """
+        dset = self.f.create_dataset('foo', shape=(100,), chunks=10)
+        self.assertEqual(dset.chunks, (10,))
+
     def test_chunks_mismatch(self):
         """ Illegal chunk size raises ValueError """
         with self.assertRaises(ValueError):
             self.f.create_dataset('foo', shape=(100,), chunks=(200,))
 
+    def test_chunks_false(self):
+        """ Chunked format required for given storage options """
+        with self.assertRaises(ValueError):
+            self.f.create_dataset('foo', shape=(10,), maxshape=100, chunks=False)
+
     def test_chunks_scalar(self):
         """ Attempting to create chunked scalar dataset raises TypeError """
         with self.assertRaises(TypeError):
@@ -236,6 +409,30 @@ class TestCreateChunked(BaseDataset):
         dset = self.f.create_dataset('foo', shape=(3,), dtype='S100000000', chunks=True)
         self.assertEqual(dset.chunks, (1,))
 
+    def test_scalar_assignment(self):
+        """ Test scalar assignment of chunked dataset """
+        dset = self.f.create_dataset('foo', shape=(3, 50, 50),
+                                     dtype=np.int32, chunks=(1, 50, 50))
+        # test assignment of selection smaller than chunk size
+        dset[1, :, 40] = 10
+        self.assertTrue(np.all(dset[1, :, 40] == 10))
+
+        # test assignment of selection equal to chunk size
+        dset[1] = 11
+        self.assertTrue(np.all(dset[1] == 11))
+
+        # test assignment of selection bigger than chunk size
+        dset[0:2] = 12
+        self.assertTrue(np.all(dset[0:2] == 12))
+
+    def test_auto_chunks_no_shape(self):
+        """ Auto-chunking of empty datasets not allowed"""
+        with pytest.raises(TypeError, match='Empty') as err:
+            self.f.create_dataset('foo', dtype='S100', chunks=True)
+
+        with pytest.raises(TypeError, match='Empty') as err:
+            self.f.create_dataset('foo', dtype='S100', maxshape=20)
+
 
 class TestCreateFillvalue(BaseDataset):
 
@@ -269,7 +466,7 @@ class TestCreateFillvalue(BaseDataset):
         self.assertAlmostEqual(dset[4], v)
 
     def test_exc(self):
-        """ Bogus fill value raises TypeError """
+        """ Bogus fill value raises ValueError """
         with self.assertRaises(ValueError):
             dset = self.f.create_dataset('foo', (10,),
                     dtype=[('a', 'i'), ('b', 'f')], fillvalue=42)
@@ -388,6 +585,16 @@ class TestCreateLZF(BaseDataset):
         self.assertEqual(dset.compression, 'lzf')
         self.assertEqual(dset.compression_opts, None)
 
+        testdata = np.arange(100)
+        dset = self.f.create_dataset('bar', data=testdata, compression='lzf')
+        self.assertEqual(dset.compression, 'lzf')
+        self.assertEqual(dset.compression_opts, None)
+
+        self.f.flush()  # Actually write to file
+
+        readdata = self.f['bar'][()]
+        self.assertArrayEqual(readdata, testdata)
+
     def test_lzf_exc(self):
         """ Giving lzf options raises ValueError """
         with self.assertRaises(ValueError):
@@ -445,6 +652,18 @@ class TestCreateScaleOffset(BaseDataset):
         with self.assertRaises(ValueError):
             dset = self.f.create_dataset('foo', (20, 30), dtype=float, scaleoffset=True)
 
+    def test_non_integer(self):
+        """ Check when scaleoffset is negetive"""
+
+        with self.assertRaises(ValueError):
+            dset = self.f.create_dataset('foo', (20, 30), dtype=float, scaleoffset=-0.1)
+
+    def test_unsupport_dtype(self):
+        """ Check when dtype is unsupported type"""
+
+        with self.assertRaises(TypeError):
+            dset = self.f.create_dataset('foo', (20, 30), dtype=bool, scaleoffset=True)
+
     def test_float(self):
         """ Scaleoffset filter works for floating point data """
 
@@ -539,7 +758,7 @@ class TestExternal(BaseDataset):
     """
         Feature: Datasets with the external storage property
     """
-    def test_external(self):
+    def test_contents(self):
         """ Create and access an external dataset """
 
         shape = (6, 100)
@@ -553,77 +772,92 @@ class TestExternal(BaseDataset):
 
         assert dset.external is not None
 
-        # verify file was created, and size is correct
-        import os
-        statinfo = os.stat(ext_file)
-        assert statinfo.st_size == testdata.nbytes
-
-        # verify contents
+        # verify file's existence, size, and contents
         with open(ext_file, 'rb') as fid:
             contents = fid.read()
-        assert contents == testdata.tostring()
+        assert contents == testdata.tobytes()
 
-    def test_external_other(self):
-        """ Test other forms of external lists """
+    def test_name_str(self):
+        """ External argument may be a file name str only """
+
+        self.f.create_dataset('foo', (6, 100), external=self.mktemp())
+
+    def test_name_path(self):
+        """ External argument may be a file name path only """
+
+        self.f.create_dataset('foo', (6, 100),
+                              external=pathlib.Path(self.mktemp()))
+
+    def test_iter_multi(self):
+        """ External argument may be an iterable of multiple tuples """
 
-        shape = (6, 100)
         ext_file = self.mktemp()
-
-        self.f.create_dataset('foo', shape, external=ext_file)
-        self.f.create_dataset('bar', shape, external=[ext_file])
-        self.f.create_dataset('moo', shape, external=[ext_file, 0])
-        self.f.create_dataset('car', shape, external=[ext_file, 0, h5f.UNLIMITED])
-
         N = 100
-        external = [(ext_file, x * 1000, (x + 1) * 1000) for x in range(0, N)]
-        dset = self.f.create_dataset('poo', shape, external=external)
+        external = iter((ext_file, x * 1000, 1000) for x in range(N))
+        dset = self.f.create_dataset('poo', (6, 100), external=external)
         assert len(dset.external) == N
 
-    def test_external_invalid(self):
+    def test_invalid(self):
         """ Test with invalid external lists """
 
         shape = (6, 100)
         ext_file = self.mktemp()
 
-        with self.assertRaises(TypeError):
-            self.f.create_dataset('foo', shape, external=[(ext_file, 0, "h5f.UNLIMITED")])
+        for exc_type, external in [
+            (TypeError, [ext_file]),
+            (TypeError, [ext_file, 0]),
+            (TypeError, [ext_file, 0, h5f.UNLIMITED]),
+            (ValueError, [(ext_file,)]),
+            (ValueError, [(ext_file, 0)]),
+            (ValueError, [(ext_file, 0, h5f.UNLIMITED, 0)]),
+            (TypeError, [(ext_file, 0, "h5f.UNLIMITED")]),
+        ]:
+            with self.assertRaises(exc_type):
+                self.f.create_dataset('foo', shape, external=external)
 
-        with self.assertRaises(TypeError):
-            self.f.create_dataset('foo', shape, external=[(ext_file, 0, h5f.UNLIMITED, 0)])
 
 class TestAutoCreate(BaseDataset):
 
     """
         Feature: Datasets auto-created from data produce the correct types
     """
+    def assert_string_type(self, ds, cset, variable=True):
+        tid = ds.id.get_type()
+        self.assertEqual(type(tid), h5py.h5t.TypeStringID)
+        self.assertEqual(tid.get_cset(), cset)
+        if variable:
+            assert tid.is_variable_str()
 
     def test_vlen_bytes(self):
-        """ Assignment of a byte string produces a vlen ascii dataset """
+        """Assigning byte strings produces a vlen string ASCII dataset """
         self.f['x'] = b"Hello there"
-        ds = self.f['x']
-        tid = ds.id.get_type()
-        self.assertEqual(type(tid), h5py.h5t.TypeStringID)
-        self.assertTrue(tid.is_variable_str())
-        self.assertEqual(tid.get_cset(), h5py.h5t.CSET_ASCII)
+        self.assert_string_type(self.f['x'], h5py.h5t.CSET_ASCII)
+
+        self.f['y'] = [b"a", b"bc"]
+        self.assert_string_type(self.f['y'], h5py.h5t.CSET_ASCII)
+
+        self.f['z'] = np.array([b"a", b"bc"], dtype=np.object_)
+        self.assert_string_type(self.f['z'], h5py.h5t.CSET_ASCII)
 
     def test_vlen_unicode(self):
-        """ Assignment of a unicode string produces a vlen unicode dataset """
-        self.f['x'] = u"Hello there" + six.unichr(0x2034)
-        ds = self.f['x']
-        tid = ds.id.get_type()
-        self.assertEqual(type(tid), h5py.h5t.TypeStringID)
-        self.assertTrue(tid.is_variable_str())
-        self.assertEqual(tid.get_cset(), h5py.h5t.CSET_UTF8)
+        """Assigning unicode strings produces a vlen string UTF-8 dataset """
+        self.f['x'] = "Hello there" + chr(0x2034)
+        self.assert_string_type(self.f['x'], h5py.h5t.CSET_UTF8)
+
+        self.f['y'] = ["a", "bc"]
+        self.assert_string_type(self.f['y'], h5py.h5t.CSET_UTF8)
+
+        # 2D array; this only works with an array, not nested lists
+        self.f['z'] = np.array([["a", "bc"]], dtype=np.object_)
+        self.assert_string_type(self.f['z'], h5py.h5t.CSET_UTF8)
 
     def test_string_fixed(self):
         """ Assignment of fixed-length byte string produces a fixed-length
         ascii dataset """
         self.f['x'] = np.string_("Hello there")
         ds = self.f['x']
-        tid = ds.id.get_type()
-        self.assertEqual(type(tid), h5py.h5t.TypeStringID)
-        self.assertEqual(tid.get_size(), 11)
-        self.assertEqual(tid.get_cset(), h5py.h5t.CSET_ASCII)
+        self.assert_string_type(ds, h5py.h5t.CSET_ASCII, variable=False)
+        self.assertEqual(ds.id.get_type().get_size(), 11)
 
 
 class TestCreateLike(BaseDataset):
@@ -647,6 +881,41 @@ class TestCreateLike(BaseDataset):
         similar = self.f.create_dataset_like('lenovo', orig)
         self.assertEqual(0, h5py.h5g.get_objinfo(similar._id).mtime)
 
+    def test_maxshape(self):
+        """ Test when other.maxshape != other.shape """
+
+        other = self.f.create_dataset('other', (10,), maxshape=20)
+        similar = self.f.create_dataset_like('sim', other)
+        self.assertEqual(similar.shape, (10,))
+        self.assertEqual(similar.maxshape, (20,))
+
+class TestChunkIterator(BaseDataset):
+    def test_no_chunks(self):
+        dset = self.f.create_dataset("foo", ())
+        with self.assertRaises(TypeError):
+            dset.iter_chunks()
+
+    def test_1d(self):
+        dset = self.f.create_dataset("foo", (100,), chunks=(32,))
+        expected = ((slice(0,32,1),), (slice(32,64,1),), (slice(64,96,1),),
+            (slice(96,100,1),))
+        self.assertEqual(list(dset.iter_chunks()), list(expected))
+        expected = ((slice(50,64,1),), (slice(64,96,1),), (slice(96,97,1),))
+        self.assertEqual(list(dset.iter_chunks(np.s_[50:97])), list(expected))
+
+    def test_2d(self):
+        dset = self.f.create_dataset("foo", (100,100), chunks=(32,64))
+        expected = ((slice(0, 32, 1), slice(0, 64, 1)), (slice(0, 32, 1),
+        slice(64, 100, 1)), (slice(32, 64, 1), slice(0, 64, 1)),
+        (slice(32, 64, 1), slice(64, 100, 1)), (slice(64, 96, 1),
+        slice(0, 64, 1)), (slice(64, 96, 1), slice(64, 100, 1)),
+        (slice(96, 100, 1), slice(0, 64, 1)), (slice(96, 100, 1),
+        slice(64, 100, 1)))
+        self.assertEqual(list(dset.iter_chunks()), list(expected))
+
+        expected = ((slice(48, 52, 1), slice(40, 50, 1)),)
+        self.assertEqual(list(dset.iter_chunks(np.s_[48:52,40:50])), list(expected))
+
 
 class TestResize(BaseDataset):
 
@@ -660,6 +929,15 @@ class TestResize(BaseDataset):
         self.assertIsNot(dset.chunks, None)
         self.assertEqual(dset.maxshape, (20, 60))
 
+    def test_create_1D(self):
+        """ Create dataset with "maxshape" using integer maxshape"""
+        dset = self.f.create_dataset('foo', (20,), maxshape=20)
+        self.assertIsNot(dset.chunks, None)
+        self.assertEqual(dset.maxshape, (20,))
+
+        dset = self.f.create_dataset('bar', 20, maxshape=20)
+        self.assertEqual(dset.maxshape, (20,))
+
     def test_resize(self):
         """ Datasets may be resized up to maxshape """
         dset = self.f.create_dataset('foo', (20, 30), maxshape=(20, 60))
@@ -669,10 +947,17 @@ class TestResize(BaseDataset):
         dset.resize((20, 60))
         self.assertEqual(dset.shape, (20, 60))
 
+    def test_resize_1D(self):
+        """ Datasets may be resized up to maxshape using integer maxshape"""
+        dset = self.f.create_dataset('foo', 20, maxshape=40)
+        self.assertEqual(dset.shape, (20,))
+        dset.resize((30,))
+        self.assertEqual(dset.shape, (30,))
+
     def test_resize_over(self):
-        """ Resizing past maxshape triggers ValueError """
+        """ Resizing past maxshape triggers an exception """
         dset = self.f.create_dataset('foo', (20, 30), maxshape=(20, 60))
-        with self.assertRaises(ValueError):
+        with self.assertRaises(Exception):
             dset.resize((20, 70))
 
     def test_resize_nonchunked(self):
@@ -783,7 +1068,7 @@ class TestStrings(BaseDataset):
         string_info = h5py.check_string_dtype(ds.dtype)
         self.assertEqual(string_info.encoding, 'utf-8')
 
-    def test_fixed_bytes(self):
+    def test_fixed_ascii(self):
         """ Fixed-length bytes dataset maps to fixed-length ascii in the file
         """
         dt = np.dtype("|S10")
@@ -797,6 +1082,22 @@ class TestStrings(BaseDataset):
         self.assertEqual(string_info.encoding, 'ascii')
         self.assertEqual(string_info.length, 10)
 
+    def test_fixed_utf8(self):
+        dt = h5py.string_dtype(encoding='utf-8', length=5)
+        ds = self.f.create_dataset('x', (100,), dtype=dt)
+        tid = ds.id.get_type()
+        self.assertEqual(tid.get_cset(), h5py.h5t.CSET_UTF8)
+        s = 'cù'
+        ds[0] = s.encode('utf-8')
+        ds[1] = s
+        ds[2:4] = [s, s]
+        ds[4:6] = np.array([s, s], dtype=object)
+        ds[6:8] = np.array([s.encode('utf-8')] * 2, dtype=dt)
+        with self.assertRaises(TypeError):
+            ds[8:10] = np.array([s, s], dtype='U')
+
+        np.testing.assert_array_equal(ds[:8], np.array([s.encode('utf-8')] * 8, dtype='S'))
+
     def test_fixed_unicode(self):
         """ Fixed-length unicode datasets are unsupported (raise TypeError) """
         dt = np.dtype("|U10")
@@ -814,17 +1115,6 @@ class TestStrings(BaseDataset):
         self.assertEqual(type(out), bytes)
         self.assertEqual(out, data)
 
-    def test_roundtrip_vlen_unicode(self):
-        """ Writing and reading to unicode dataset preserves type and content
-        """
-        dt = h5py.string_dtype()
-        ds = self.f.create_dataset('x', (100,), dtype=dt)
-        data = u"Hello" + six.unichr(0x2034)
-        ds[0] = data
-        out = ds[0]
-        self.assertEqual(type(out), six.text_type)
-        self.assertEqual(out, data)
-
     def test_roundtrip_fixed_bytes(self):
         """ Writing to and reading from fixed-length bytes dataset preserves
         type and content """
@@ -836,14 +1126,64 @@ class TestStrings(BaseDataset):
         self.assertEqual(type(out), np.string_)
         self.assertEqual(out, data)
 
-    @ut.expectedFailure
-    def test_unicode_write_error(self):
-        """ Writing a non-utf8 byte string to a unicode vlen dataset raises
-        ValueError """
+    def test_retrieve_vlen_unicode(self):
         dt = h5py.string_dtype()
+        ds = self.f.create_dataset('x', (10,), dtype=dt)
+        data = "fàilte"
+        ds[0] = data
+        self.assertIsInstance(ds[0], bytes)
+        out = ds.asstr()[0]
+        self.assertIsInstance(out, str)
+        self.assertEqual(out, data)
+
+    def test_asstr(self):
+        ds = self.f.create_dataset('x', (10,), dtype=h5py.string_dtype())
+        data = "fàilte"
+        ds[0] = data
+
+        strwrap1 = ds.asstr('ascii')
+        with self.assertRaises(UnicodeDecodeError):
+            out = strwrap1[0]
+
+        # Different errors parameter
+        self.assertEqual(ds.asstr('ascii', 'ignore')[0], 'filte')
+
+        # latin-1 will decode it but give the wrong text
+        self.assertNotEqual(ds.asstr('latin-1')[0], data)
+
+        # Array output
+        np.testing.assert_array_equal(
+            ds.asstr()[:1], np.array([data], dtype=object)
+        )
+
+    def test_asstr_fixed(self):
+        dt = h5py.string_dtype(length=5)
+        ds = self.f.create_dataset('x', (10,), dtype=dt)
+        data = 'cù'
+        ds[0] = np.array(data.encode('utf-8'), dtype=dt)
+
+        self.assertIsInstance(ds[0], np.bytes_)
+        out = ds.asstr()[0]
+        self.assertIsInstance(out, str)
+        self.assertEqual(out, data)
+
+        # Different errors parameter
+        self.assertEqual(ds.asstr('ascii', 'ignore')[0], 'c')
+
+        # latin-1 will decode it but give the wrong text
+        self.assertNotEqual(ds.asstr('latin-1')[0], data)
+
+        # Array output
+        np.testing.assert_array_equal(
+            ds.asstr()[:1], np.array([data], dtype=object)
+        )
+
+    def test_unicode_write_error(self):
+        """Encoding error when writing a non-ASCII string to an ASCII vlen dataset"""
+        dt = h5py.string_dtype('ascii')
         ds = self.f.create_dataset('x', (100,), dtype=dt)
-        data = "Hello\xef"
-        with self.assertRaises(ValueError):
+        data = "fàilte"
+        with self.assertRaises(UnicodeEncodeError):
             ds[0] = data
 
     def test_unicode_write_bytes(self):
@@ -851,12 +1191,23 @@ class TestStrings(BaseDataset):
         """
         dt = h5py.string_dtype()
         ds = self.f.create_dataset('x', (100,), dtype=dt)
-        data = u"Hello there" + six.unichr(0x2034)
-        ds[0] = data.encode('utf8')
+        data = (u"Hello there" + chr(0x2034)).encode('utf8')
+        ds[0] = data
         out = ds[0]
-        self.assertEqual(type(out), six.text_type)
+        self.assertEqual(type(out), bytes)
         self.assertEqual(out, data)
 
+    def test_vlen_bytes_write_ascii_str(self):
+        """ Writing an ascii str to ascii vlen dataset is OK
+        """
+        dt = h5py.string_dtype('ascii')
+        ds = self.f.create_dataset('x', (100,), dtype=dt)
+        data = "ASCII string"
+        ds[0] = data
+        out = ds[0]
+        self.assertEqual(type(out), bytes)
+        self.assertEqual(out, data.encode('ascii'))
+
 
 class TestCompound(BaseDataset):
 
@@ -901,6 +1252,46 @@ class TestCompound(BaseDataset):
         self.assertTrue(np.all(outdata == testdata))
         self.assertEqual(outdata.dtype, testdata.dtype)
 
+    def test_fields(self):
+        dt = np.dtype([
+            ('x', np.float64),
+            ('y', np.float64),
+            ('z', np.float64),
+        ])
+
+        testdata = np.ndarray((16,), dtype=dt)
+        for key in dt.fields:
+            testdata[key] = np.random.random((16,)) * 100
+
+        self.f['test'] = testdata
+
+        # Extract multiple fields
+        np.testing.assert_array_equal(
+            self.f['test'].fields(['x', 'y'])[:], testdata[['x', 'y']]
+        )
+        # Extract single field
+        np.testing.assert_array_equal(
+            self.f['test'].fields('x')[:], testdata['x']
+        )
+
+
+class TestSubarray(BaseDataset):
+    def test_write_list(self):
+        ds = self.f.create_dataset("a", (1,), dtype="3int8")
+        ds[0] = [1, 2, 3]
+        np.testing.assert_array_equal(ds[:], [[1, 2, 3]])
+
+        ds[:] = [[4, 5, 6]]
+        np.testing.assert_array_equal(ds[:], [[4, 5, 6]])
+
+    def test_write_array(self):
+        ds = self.f.create_dataset("a", (1,), dtype="3int8")
+        ds[0] = np.array([1, 2, 3])
+        np.testing.assert_array_equal(ds[:], [[1, 2, 3]])
+
+        ds[:] = np.array([[4, 5, 6]])
+        np.testing.assert_array_equal(ds[:], [[4, 5, 6]])
+
 
 class TestEnum(BaseDataset):
 
@@ -967,6 +1358,11 @@ class TestTrackTimes(BaseDataset):
         ds_mtime = h5py.h5g.get_objinfo(ds._id).mtime
         self.assertEqual(0, ds_mtime)
 
+    def test_invalid_track_times(self):
+        """ check that when give track_times an invalid value """
+        with self.assertRaises(TypeError):
+            self.f.create_dataset('foo', (4,), track_times='null')
+
 
 class TestZeroShape(BaseDataset):
 
@@ -976,10 +1372,13 @@ class TestZeroShape(BaseDataset):
 
     def test_array_conversion(self):
         """ Empty datasets can be converted to NumPy arrays """
-        ds = self.f.create_dataset('x', (0,), maxshape=(None,))
+        ds = self.f.create_dataset('x', 0, maxshape=None)
         self.assertEqual(ds.shape, np.array(ds).shape)
 
-        ds = self.f.create_dataset('y', (0, 0), maxshape=(None, None))
+        ds = self.f.create_dataset('y', (0,), maxshape=(None,))
+        self.assertEqual(ds.shape, np.array(ds).shape)
+
+        ds = self.f.create_dataset('z', (0, 0), maxshape=(None, None))
         self.assertEqual(ds.shape, np.array(ds).shape)
 
     def test_reading(self):
@@ -993,6 +1392,11 @@ class TestZeroShape(BaseDataset):
         self.assertEqual(ds[()].shape, arr.shape)
         self.assertEqual(ds[()].dtype, arr.dtype)
 
+# https://github.com/h5py/h5py/issues/1492
+empty_regionref_xfail = pytest.mark.xfail(
+    h5py.version.hdf5_version_tuple == (1, 10, 6),
+    reason="Issue with empty region refs in HDF5 1.10.6",
+)
 
 class TestRegionRefs(BaseDataset):
 
@@ -1012,6 +1416,28 @@ class TestRegionRefs(BaseDataset):
         ref = self.dset.regionref[slic]
         self.assertArrayEqual(self.dset[ref], self.data[slic])
 
+    @empty_regionref_xfail
+    def test_empty_region(self):
+        ref = self.dset.regionref[:0]
+        out = self.dset[ref]
+        assert out.size == 0
+        # Ideally we should preserve shape (0, 100), but it seems this is lost.
+
+    @empty_regionref_xfail
+    def test_scalar_dataset(self):
+        ds = self.f.create_dataset("scalar", data=1.0, dtype='f4')
+        sid = h5py.h5s.create(h5py.h5s.SCALAR)
+
+        # Deselected
+        sid.select_none()
+        ref = h5py.h5r.create(ds.id, b'.', h5py.h5r.DATASET_REGION, sid)
+        assert ds[ref] == h5py.Empty(np.dtype('f4'))
+
+        # Selected
+        sid.select_all()
+        ref = h5py.h5r.create(ds.id, b'.', h5py.h5r.DATASET_REGION, sid)
+        assert ds[ref] == ds[()]
+
     def test_ref_shape(self):
         """ Region reference shape and selection shape """
         slic = np.s_[25:35, 10:100:5]
@@ -1021,19 +1447,22 @@ class TestRegionRefs(BaseDataset):
 
 
 class TestAstype(BaseDataset):
-
+    """.astype() wrapper & context manager
     """
-        .astype context manager
-    """
-
-    def test_astype(self):
-
+    def test_astype_ctx(self):
         dset = self.f.create_dataset('x', (100,), dtype='i2')
         dset[...] = np.arange(100)
         with dset.astype('f8'):
-            self.assertEqual(dset[...].dtype, np.dtype('f8'))
-            self.assertTrue(np.all(dset[...] == np.arange(100)))
+            self.assertArrayEqual(dset[...], np.arange(100, dtype='f8'))
 
+        with dset.astype('f4') as f4ds:
+            self.assertArrayEqual(f4ds[...], np.arange(100, dtype='f4'))
+
+    def test_astype_wrapper(self):
+        dset = self.f.create_dataset('x', (100,), dtype='i2')
+        dset[...] = np.arange(100)
+        arr = dset.astype('f4')[:]
+        self.assertArrayEqual(arr, np.arange(100, dtype='f4'))
 
 class TestScalarCompound(BaseDataset):
 
@@ -1061,7 +1490,7 @@ class TestVlen(BaseDataset):
         self.assertArrayEqual(ds[1], np.arange(0))
         self.assertArrayEqual(ds[2], np.array([1, 2, 3]))
         self.assertArrayEqual(ds[1], np.arange(0))
-        ds[0:2] = np.array([np.arange(5), np.arange(4)])
+        ds[0:2] = np.array([np.arange(5), np.arange(4)], dtype=object)
         self.assertArrayEqual(ds[0], np.arange(5))
         self.assertArrayEqual(ds[1], np.arange(4))
         ds[0:2] = np.array([np.arange(3), np.arange(3)])
@@ -1090,7 +1519,7 @@ class TestVlen(BaseDataset):
         self.assertArrayEqual(ds[0], np.array([1, 1]))
         self.assertArrayEqual(ds[1], np.array([1]))
         self.assertArrayEqual(ds[2], np.array([1, 2, 3]))
-        ds[0:2] = np.array([[0.1, 1.1, 2.1, 3.1, 4], np.arange(4)])
+        ds[0:2] = np.array([[0.1, 1.1, 2.1, 3.1, 4], np.arange(4)], dtype=object)
         self.assertArrayEqual(ds[0], np.arange(5))
         self.assertArrayEqual(ds[1], np.arange(4))
         ds[0:2] = np.array([np.array([0.1, 1.2, 2.2]),
@@ -1103,7 +1532,7 @@ class TestVlen(BaseDataset):
         ds = self.f.create_dataset('vlen', (2, 2), dtype=dt)
         ds[0, 0] = np.arange(1)
         ds[:, :] = np.array([[np.arange(3), np.arange(2)],
-                            [np.arange(1), np.arange(2)]])
+                            [np.arange(1), np.arange(2)]], dtype=object)
         ds[:, :] = np.array([[np.arange(2), np.arange(2)],
                              [np.arange(2), np.arange(2)]])
 
@@ -1171,6 +1600,20 @@ class TestVlen(BaseDataset):
         np_dt = np.float64
         self._help_float_testing(np_dt)
 
+    def test_non_contiguous_arrays(self):
+        """Test that non-contiguous arrays are stored correctly"""
+        self.f.create_dataset('nc', (10,), dtype=h5py.vlen_dtype('bool'))
+        x = np.array([True, False, True, True, False, False, False])
+        self.f['nc'][0] = x[::2]
+
+        assert all(self.f['nc'][0] == x[::2]), f"{self.f['nc'][0]} != {x[::2]}"
+
+        self.f.create_dataset('nc2', (10,), dtype=h5py.vlen_dtype('int8'))
+        y = np.array([2, 4, 1, 5, -1, 3, 7])
+        self.f['nc2'][0] = y[::2]
+
+        assert all(self.f['nc2'][0] == y[::2]), f"{self.f['nc2'][0]} != {y[::2]}"
+
 
 class TestLowOpen(BaseDataset):
 
@@ -1186,3 +1629,98 @@ class TestLowOpen(BaseDataset):
         del dset
         dsid = h5py.h5d.open(self.f.id, b'x', dapl)
         self.assertIsInstance(dsid, h5py.h5d.DatasetID)
+
+
+@ut.skipUnless(h5py.version.hdf5_version_tuple >= (1, 10, 5),
+               "chunk info requires  HDF5 >= 1.10.5")
+def test_get_chunk_details():
+    from io import BytesIO
+    buf = BytesIO()
+    with h5py.File(buf, 'w') as fout:
+        fout.create_dataset('test', shape=(100, 100), chunks=(10, 10), dtype='i4')
+        fout['test'][:] = 1
+
+    buf.seek(0)
+    with h5py.File(buf, 'r') as fin:
+        ds = fin['test'].id
+
+        assert ds.get_num_chunks() == 100
+        for j in range(100):
+            offset = tuple(np.array(np.unravel_index(j, (10, 10))) * 10)
+
+            si = ds.get_chunk_info(j)
+            assert si.chunk_offset == offset
+            assert si.filter_mask == 0
+            assert si.byte_offset is not None
+            assert si.size > 0
+
+        si = ds.get_chunk_info_by_coord((0, 0))
+        assert si.chunk_offset == (0, 0)
+        assert si.filter_mask == 0
+        assert si.byte_offset is not None
+        assert si.size > 0
+
+
+def test_empty_shape(writable_file):
+    ds = writable_file.create_dataset('empty', dtype='int32')
+    assert ds.shape is None
+    assert ds.maxshape is None
+
+
+def test_zero_storage_size():
+    # https://github.com/h5py/h5py/issues/1475
+    from io import BytesIO
+    buf = BytesIO()
+    with h5py.File(buf, 'w') as fout:
+        fout.create_dataset('empty', dtype='uint8')
+
+    buf.seek(0)
+    with h5py.File(buf, 'r') as fin:
+        assert fin['empty'].chunks is None
+        assert fin['empty'].id.get_offset() is None
+        assert fin['empty'].id.get_storage_size() == 0
+
+
+def test_python_int_uint64(writable_file):
+    # https://github.com/h5py/h5py/issues/1547
+    data = [np.iinfo(np.int64).max, np.iinfo(np.int64).max + 1]
+
+    # Check creating a new dataset
+    ds = writable_file.create_dataset('x', data=data, dtype=np.uint64)
+    assert ds.dtype == np.dtype(np.uint64)
+    np.testing.assert_array_equal(ds[:], np.array(data, dtype=np.uint64))
+
+    # Check writing to an existing dataset
+    ds[:] = data
+    np.testing.assert_array_equal(ds[:], np.array(data, dtype=np.uint64))
+
+
+def test_setitem_fancy_indexing(writable_file):
+    # https://github.com/h5py/h5py/issues/1593
+    arr = writable_file.create_dataset('data', (5, 1000, 2), dtype=np.uint8)
+    block = np.random.randint(255, size=(5, 3, 2))
+    arr[:, [0, 2, 4], ...] = block
+
+
+def test_vlen_spacepad():
+    with File(get_data_file_path("vlen_string_dset.h5")) as f:
+        assert f["DS1"][0] == b"Parting"
+
+
+def test_vlen_nullterm():
+    with File(get_data_file_path("vlen_string_dset_utc.h5")) as f:
+        assert f["ds1"][0] == b"2009-12-20T10:16:18.662409Z"
+
+
+@pytest.mark.skipif(
+    h5py.version.hdf5_version_tuple < (1, 10, 3),
+    reason="Appears you cannot pass an unknown filter id for HDF5 < 1.10.3"
+)
+def test_allow_unknown_filter(writable_file):
+    # apparently 256-511 are reserved for testing purposes
+    fake_filter_id = 256
+    ds = writable_file.create_dataset(
+        'data', shape=(10, 10), dtype=np.uint8, compression=fake_filter_id,
+        allow_unknown_filter=True
+    )
+    assert str(fake_filter_id) in ds._filters
